import numpy as np
import pandas as pd


class PocotoTest():

    def __init__(self, env='dev'):
        self.name = 'Test'
        self.rapport = {}

    def set_name(self, name):
        self.name = name
        self.rapport[name] = {}

    def display_rapport(self):
        for key, value in self.rapport.items():
            print('\n')
            print('>', key)
            print(value)

    def check_missing_values(self, df):
        tmp = df.isnull().sum()
        tmp = tmp[tmp > 0].to_dict()

        print('\n> Checking missing values')
        for key, value in tmp.items():
            print('MISSING', value, '\t', key)
        if len(tmp.keys()) == 0:
            print('- No missing values')
            df['missing'] = None
        self.rapport[self.name]['missing'] = np.NaN
        return df

    def check_min_max_values(self, df, list_of_col):
        print('\n> Checking missing values')
        for col in list_of_col:
            print(col, '-' 'min :\t', df[col].min())
            print(col, '-' 'max :\t', df[col].max())
        return df

    def get_range(self, df, col='datetime'):
        return df[col].max() - df[col].min()

    def check_datetime_range(self, df):
        start = df.datetime.min()
        end = df.datetime.max()
        range_ = self.get_range(df)
        print('\n> Time information')
        print('START :\t', start)
        print('END   :\t', end)
        print('Total :\t', range_)
        return df

    def check_four_information(self, df):
        print('\n> Values per furnace')
        tmp = df['four'].value_counts().sort_index()
        for key, value in tmp.items():
            print('Four', '\t', key, '\t', value)
        return df

    def check_information(self, df, col='four', surname='Four'):
        print('\n> Values per {}'.format(surname))
        tmp = df[col].value_counts().sort_index()
        for key, value in tmp.items():
            print(surname, '\t', key, '\t', value)
        return df

    def check_four_information(self, df):
        return self.check_information(df, 'four', 'Four')

    def check_type_information(self, df):
        return self.check_information(df, 'type', 'Type')

    def check_face_information(self, df):
        return self.check_information(df, 'face', 'Face')

    def check_tag_information(self, df):
        range_ = self.get_range(df)
        
        # Checking Tag informations
        print('\n> Values per Tag')
        nb_tag = df['tag'].nunique()
        val_per_tag_real = df.groupby('tag').size().mean().round()
        val_per_tag_supo = np.round(range_ / pd.Timedelta('10min'), 0)
        pct_per_tag = np.round((val_per_tag_real / val_per_tag_supo) * 100, 2)
        print('Tags nb\t', nb_tag)
        print(val_per_tag_real, '\t(avg value per tag)')
        print(val_per_tag_supo, '\t(supposed value per tag)')
        print(pct_per_tag, '\t%')

        # Checking errors in description
        tmp = df.groupby('tag').nunique()['tag']
        tmp = tmp[tmp != 1]
        tmp = tmp[tmp > 0].to_dict()
        for key, value in tmp.items():
            print('TAG - ', key, '\t', value, 'different descriptions')
        return df

    def check_alerts(self, df):
        print('\n> Checking alerts')
        tmp = df['alert_num'].value_counts()
        for key, value in tmp.items():
            pct_value = np.round((value / df.shape[0]) * 100, 2)
            print('Alerts', '\t', key, '\t', pct_value, '%')
        return df

    def print_test(self):
        print('Test', self.name)



# Making all necessary python imports for library use
import random as __random
import numpy as __np
import pandas as __pd
from PIL import Image as __Image  # pip install pillow
# from statsmodels.tsa.stattools import adfuller


from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

"""
#################################################################################

List of functions :

 - create_clean_df_from_cloud_json
 - create_image
 - json_to_df
 - set_timestamp_df_index

#################################################################################
"""


def create_clean_df_from_cloud_json(js):
    """ Given a json downloaded on the cloud from get_datasource_data or get_location_data, returns a corrected df"""
    df = json_to_df(js)
    df_data = set_timestamp_df_index(df)
    for col in df_data.columns:
        if col[:6] == 'values':
            df_data[col] = __pd.to_numeric(df_data[col], errors='coerce')
    return df_data


def create_image():
    size = (256, 256)
    im = __Image.new('RGB', size)
    pix = im.load()
    for i in range(size[0]):
        for j in range(size[1]):
            pix[i, j] = (i, int(0.5 * i + __random.randint(0, 127)), int(0.5 * j + __random.randint(0, 127)))
    im.show()


def json_to_df(json_file, keys=False):
    """Description : Transforms data of a json to a pandas DataFrame
    Takes : json_file as list of dictionnaries
    Returns : A pandas dataFrame with all the datas converted in nice columns easy to deal with
    """
    try:
        # lists_initialization
        liste_keys = []
        new_list = []

        # still_dic is a variable which detects if there is still a deeper dictionnary into the current dictionnary
        still_dic = True
        compteur = 0

        # While there is a dictionnary we go down into it to get #order1 dictionnary
        while still_dic == True:
            # still_dic is set to false and if we find a dictionnary we'll set it back to True
            still_dic = False
            # json_file is the list of dictionnaries in the json_file : let's read it
            # We want to detect the dictionnaries odf deeper order and transform them into simple dictionnaries

            for bloc in json_file:
                new_dic = {}
                # print(bloc)
                for i in bloc.keys():
                    if type(bloc[i]) == dict:
                        for j in bloc[i].keys():
                            # renaming {key_1:{key_2:'bla'}} in {key_1_key_2:'bla'}
                            liste_keys.append(str(i) + '_' + str(j))
                            new_dic[str(i) + '_' + str(j)] = bloc[i][j]
                            if type(bloc[i][j]) == dict:
                                still_dic = True
                    else:
                        liste_keys.append(i)
                        new_dic[str(i)] = bloc[i]
                new_list.append(new_dic)
                # Once we have transformed dictionnaries in dictionaries to simple dictionnaries with more keys we remplace the json file
            # It will still have the same length

            json_file = new_list
            new_list = []
            compteur += 1

        # We now have a json file with n dictionnaries of order one (no dict in dict)
        # getting all its keys in alphabetical order
        liste_keys = list(set(liste_keys))
        liste_keys.sort()
        final = []

        # looping on the json_file we transform dict in columns and dataFrame
        for i in json_file:
            interm = []
            for elt in liste_keys:
                interm.append(i[elt]) if elt in i else interm.append('')
            final.append(interm)
        data_array = __np.array(final)
        df_data = __pd.DataFrame(data_array, columns=liste_keys)
        # print('Converting json file to df : End')

        if keys:
            return df_data, liste_keys
        else:
            return df_data
    except:
        print('Error : DataFrame couldn\'t be transformed from json')


def set_timestamp_df_index(df, column='timestamp', del_col=True):
    """used to transform dataFrame with cloud timestamp format (%Y-%m-%dT%H:%M:%S.000Z)
    into dataFrame with pandas timestamp index.
    del_col (true by default) : delete the column with the original timestamp format
    column : name of the column to deal with.
    Returns the corrected dataFrame"""
    init_df = df.copy()
    try:
        df[column] = __pd.to_numeric(df[column])
        df[column] = __pd.to_datetime(df[column].apply(t.format_to_df_time))
        df.index = df[column]
        if del_col:
            del (df[column])
        return df

    except:
        return init_df


# def test_stationarity(timeseries, window=12):
#     # Determing rolling statistics
#     rolmean = timeseries.rolling(window=window).mean()
#     rolstd = timeseries.rolling(window=window).std()

#     # Plot rolling statistics:
#     orig = plt.plot(timeseries, color='blue', label='Original')
#     mean = plt.plot(rolmean, color='red', label='Rolling Mean')
#     std = plt.plot(rolstd, color='black', label='Rolling Std')
#     plt.legend(loc='best')
#     plt.title('Rolling Mean & Standard Deviation')
#     plt.show(block=False)

#     # Perform Dickey-Fuller test:
#     print('Results of Dickey-Fuller Test:')
#     dftest = adfuller(timeseries, autolag='AIC')
#     dfoutput = pd.Series(dftest[0:4], index=['Test Statistic', 'p-value', '#Lags Used', 'Number of Observations Used'])
#     for key, value in dftest[4].items():
#         dfoutput['Critical Value (%s)' % key] = value
#     print(dfoutput)


def get_granularity(start_date, end_date):
    range_list = [(365, 'year'), (90, '3months'), (30, 'month'), (7, 'week'), (1, 'day')]

    diff_days = (end_date - start_date).days
    if diff_days < 0:
        return 'Error : start_date ({}) < end_date ({})'.format(start_date, end_date)
    for i in range_list:
        if diff_days >= i[0]:
            return i[1]
    return range_list[-1][1]


if __name__ == "__main__":
    print("All secundary functions have been loaded.")



# -*- coding: utf-8 -*-

#################################################################################
###                                                                           ###
###                                                                           ###
###                                     GENERAL                               ###
###                                                                           ###
###                                                                           ###
#################################################################################


# Making all necessary python imports for library use
import os as __os
import scipy as __sp
import numpy as __np
import math as __math
import json as __json
import pandas as __pd
import string as __string
import logging as __logging
import datetime as __datetime
import configparser as __configparser
from IPython.core.display import display, HTML

# Making local imports
# import time_format as t

# This is to avoid unuseful python warning with pandas
__pd.options.mode.chained_assignment = None

""""
#################################################################################
List of functions : 

 - (i) line
 - (i) parse_json
 - (i) url_to_json
 - (i) write_to_json_file
 - (i) read_file
 - (i) get_time_range_info_from_df
 - (i) save_df_to_excel
 - (i) transfoCol
 - (i) renameDfCol
 - (i) getPctValue
 - (i) star_print
 - (i) get_functions_in_file
 - (i) spacify_number

Indicators of importance :

(i)    : Important functions
(ii)   : Intermediary used functions
(iii)  : Not really used
(iii) !: To be deleted
 
#################################################################################
"""


def line(display=False):
    """To make easy separations while printing in console"""
    if display == True:
        return print("____________________________________________________________\n")



def num_to_str(num):
    """Function to convert a float to a string in dataframe
    (Used for dataframe => Might be deleted from this file"""
    return str(num)


def flatten_json(nested_json):
    """
        Flatten json object with nested keys into a single level.
        Args:
            nested_json: A nested json object.
        Returns:
            The flattened json object if successful, None otherwise.
    """
    out = {}

    def flatten(x, name=''):
        if type(x) is dict:
            for a in x:
                flatten(x[a], name + a + '_')
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + str(i) + '_')
                i += 1
        else:
            out[name[:-1]] = x

    flatten(nested_json)
    return out

# def get_time_range_info_from_df(df, display=True, datetime_col=None):
#     if not datetime_col:
#         start = str(df.index[0])
#         end = str(df.index[-1])
#     else:
#         start = str(df[datetime_col].iloc[0])
#         end = str(df[datetime_col].iloc[-1])
#     if display:
#         print('Start : ', start, '\t', t.format_to_numeric_timestamp(start))
#         print('End   : ', end, '\t', t.format_to_numeric_timestamp(end))
#     return start, end


def save_df_to_excel(df, name='my_df_file.xlsx', open_=False):
    """This functions takes a dataFrame or list od dataFrames ans save it into an excel with one or multiple sheets.
    Parameters : df (dataframe to be saved), name (the name of the excel file with .xlsx at the end of it)
    """

    # If df is a single dataFrame : saving it into a classic excel file with one sheet
    if type(df) == __pd.core.frame.DataFrame:
        try:
            df.to_excel(name)
            print('File : ' + str(name) + ' has correctly been saved in the current folder')
            if open_:
                cmd_ = 'start excel.exe ' + name
                __os.system(cmd_)
        except:
            print('File : ' + str(name) + ' couldn\'t be saved. Might be open. Re-try.')
            # Check errors : could be because xlsx was not added

    # If df is a list of dataFrames : saving each of them in a sheet
    elif type(df) == list:
        list_df = df
        try:
            writer = __pd.ExcelWriter(name)
            compteur = 0
            for i in list_df:
                if len(i) == 1:
                    i[1].to_excel(writer, "Sheet_" + compteur)
                    compteur += 1
                if len(i) == 2:
                    i[1].to_excel(writer, i[0])
            writer.save()
            print('File : ' + str(name) + ' has correctly been saved in the current folder as a list of dataFrames')
        except:
            print('File : ' + str(name) + ' couldn\'t be saved as a list of dataFrames. Might be open. Re-try.')
            # Check errors : could be because xlsx was not added

    # If df is not one of the two defined formats => Error
    else:
        print('File send was not a dataFrame neither')
        # Add an assertion error ?



def transfo_col(ancien, ponctuation=None, accent=None, replacer='_'):
    """Description :
        simplifie une chaine de caractère en supprimant les majuscules, la ponctuation, les accents et les espaces
    inputs :
        - ancien as string : chaine à modifier
        - ponctuation as list : liste des caractères à retirer
        - accent as dict : dictionnaire des caractères à modifier par un autre
    outputs:
        - string : chaine de caractère modifiée (simplifiée)
    """

    if not ponctuation:
        caracters_to_remove = list(__string.punctuation) + [' ', '°']
        ponctuation = {initial: replacer for initial in caracters_to_remove}

    if not accent:
        avec_accent = ['é', 'è', 'ê', 'à', 'ù', 'ç', 'ô', 'î', 'â']
        sans_accent = ['e', 'e', 'e', 'a', 'u', 'c', 'o', 'i', 'a']
        accent = {sans: avec for sans, avec in zip(avec_accent, sans_accent)}

    ancien = str(ancien)
    ancien = ancien.lower()
    ancien = ancien.translate(str.maketrans(ponctuation))
    ancien = ancien.translate(str.maketrans(accent))
    double_replacer = replacer + replacer
    while double_replacer in ancien:
        ancien = ancien.replace(double_replacer, replacer)

    if ancien[0] == replacer:
        ancien = ancien[1:]

    if ancien[-1] == replacer:
        ancien = ancien[:-1]

    return ancien


def rename_df_col(df, replacer='_'):
    """Description : uniformise le nom des colonnes d'un dataframe en retirant les caractères spéciaux/surabondants
    inputs :
        - df as dataFrame : tableau de données dont les colonnes sont à renommer de manière plus simple
    outputs:
        - dataFrame : tableau de données dont les noms de colonnes ont été modifiés
    """
    rename_dict = {ancien: transfo_col(ancien, replacer=replacer) for ancien in df.columns}
    df_new = df.rename(columns=rename_dict)
    return df_new


def get_pct_value(value_, text_=False):
    pct = __np.round(value_ * 100, 2)
    if text_:
        return '{} %'.format(pct)
    return pct


def star_print(text, stars=10, length=None, symbol='*'):
    """Affichage d'une ligne avec une valeur au milieu
    Nécessite un texte, variable text as string
    Choix du nombre de caractères, variable length as int
    Choix du nombre d'étoiles avant/après le texte, variable stars as int
    Choix du symbole, variable symbol as string
    """
    if not length:
        return print(symbol * stars, text, symbol * stars)

    text_len = len(text)
    if text_len > length:
        return print(symbol * stars, text, symbol * stars)

    stars_start = ((length - text_len) / 2) - 1
    if stars_start == int(stars_start):
        return print(symbol * int(stars_start) + ' ' + text + ' ' + symbol * int(stars_start))
    else:
        stars_start = int(stars_start)
        return print(symbol * stars_start + ' ' + text + ' ' + symbol * (stars_start + 1))


def get_functions_in_file(file_name, print_=False):
    my_file = read_file(file_name=file_name)
    sp = my_file.split('\n')
    def_line = [i[4:].split('(')[0] for i in sp if i[:3] == 'def']

    if print_:
        print('List of functions')
        for i in def_line:
            print(' - {}'.format(i))
    return def_line


def get_categorical(df, verbose=False):
    cat_data = df.select_dtypes(include='object')
    num_data = df.select_dtypes(exclude='object')

    cat_cols = cat_data.columns.values
    num_cols = num_data.columns.values

    if verbose:
        print('\nCategorical Columns :\n ', cat_cols)
        print('\nNumerical Columns : \n', num_cols)
    return cat_cols, num_cols


def get_pct_empty_df(df):
    return __pd.DataFrame({'Emptiness (%)': __np.round((df.isnull().sum() / df.shape[0]) * 100)})


def return_middle_date(date1, date2):
    if date1 > date2:
        new_date = date2 + (date1 - date2) / 2
    else:
        new_date = date1 + (date2 - date1) / 2
    return new_date


def display_html(rep):
    display(HTML(rep.text))


def get_config(config_path):
    config = __configparser.ConfigParser()
    config.read(config_path)
    return config


def millify(n):
    n = float(n)
    millnames = ["", " K", " M", " B", " T"]
    millidx = max(0, min(len(millnames) - 1, int(__math.floor(0 if n == 0 else __math.log10(abs(n)) / 3))), )
    return "{:.0f}{}".format(n / 10 ** (3 * millidx), millnames[millidx])


def read_jl_file(file_name):
    values = []
    with open(file_name, 'rb') as file:
        line_ = 'line'
        while len(line_) > 1:
            line_ = file.readline()
            values.append(line_)
    values = values[:-1]
    values = [__json.loads(i) for i in values]
    df = __pd.DataFrame(values)
    return df


def save_ts_analyse():
    ts = str(__datetime.datetime.now())[:19]
    with open('data/last_analyse.txt', 'w') as file:
        file.write(ts)


def load_ts_analyse():
    with open('data/last_analyse.txt', 'r') as file:
        ts = file.read()
    ts = __pd.to_datetime(ts)
    return ts


def get_now():
    return str(__datetime.datetime.now())[:19].replace(' ', '_').replace(':', '-')


def define_logger():
    """ Example of use :
    logger = f.define_logger()
    logger.info('Start')
    """
    # create logger
    logger = __logging.getLogger('log')
    logger.setLevel(__logging.DEBUG)

    # create console handler and set level to debug
    ch = __logging.StreamHandler()
    ch.setLevel(__logging.DEBUG)

    # create formatter
    formatter = __logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)
    return logger


def timer(start_time=None):
    if not start_time:
        start_time = __datetime.datetime.now()
        return start_time
    elif start_time:
        thour, temp_sec = divmod((__datetime.datetime.now() - start_time).total_seconds(), 3600)
        tmin, tsec = divmod(temp_sec, 60)
        if thour < 1:
            print('\n Time taken: %i minutes and %s seconds.' % (tmin, round(tsec, 2)))
        else:
            print('\n Time taken: %i hours %i minutes and %s seconds.' % (thour, tmin, round(tsec, 2)))


# source https://www.kaggle.com/krishonaveen/xtreme-boost-and-feature-engineering
def reduce_mem_usage(df, verbose=True):
    numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
    start_mem = df.memory_usage().sum() / 1024 ** 2
    for col in df.columns:
        col_type = df[col].dtypes
        if col_type in numerics:
            c_min = df[col].min()
            c_max = df[col].max()
            if str(col_type)[:3] == 'int':
                if c_min > __np.iinfo(__np.int8).min and c_max < __np.iinfo(__np.int8).max:
                    df[col] = df[col].astype(__np.int8)
                elif c_min > __np.iinfo(__np.int16).min and c_max < __np.iinfo(__np.int16).max:
                    df[col] = df[col].astype(__np.int16)
                elif c_min > __np.iinfo(__np.int32).min and c_max < __np.iinfo(__np.int32).max:
                    df[col] = df[col].astype(__np.int32)
                elif c_min > __np.iinfo(__np.int64).min and c_max < __np.iinfo(__np.int64).max:
                    df[col] = df[col].astype(__np.int64)
            else:
                if c_min > __np.finfo(__np.float16).min and c_max < __np.finfo(__np.float16).max:
                    df[col] = df[col].astype(__np.float16)
                elif c_min > __np.finfo(__np.float32).min and c_max < __np.finfo(__np.float32).max:
                    df[col] = df[col].astype(__np.float32)
                else:
                    df[col] = df[col].astype(__np.float64)
    end_mem = df.memory_usage().sum() / 1024 ** 2

    if verbose:
        print('Mem. usage decreased to {:5.2f} Mb ({:.1f}% reduction) - From {} Mo to {} Mo.'.format(
            end_mem,
            100 * (start_mem - end_mem) / start_mem,
            __np.round(start_mem, 2),
            __np.round(end_mem, 2))
        )
    return df


def resumetable(df):
    print(f"Dataset Shape: {df.shape}")
    summary = __pd.DataFrame(df.dtypes, columns=['dtypes'])
    summary = summary.reset_index()
    summary['Name'] = summary['index']
    summary = summary[['Name', 'dtypes']]
    summary['Missing'] = df.isnull().sum().values
    summary['Uniques'] = df.nunique().values
    summary['First Value'] = df.loc[0].values
    summary['Second Value'] = df.loc[1].values
    summary['Third Value'] = df.loc[2].values

    for name in summary['Name'].value_counts().index:
        summary.loc[summary['Name'] == name, 'Entropy'] = round(
            __sp.stats.entropy(df[name].value_counts(normalize=True), base=2), 2)

    return summary


def split_ID_number(df, nb_split=7):
    for i in range(nb_split):
        df['TransactionID_{}'.format(i)] = df['TransactionID'].apply(lambda x: int(str(x)[i]))
    return df


def create_from_dt(df, col_name='TransactionDT'):
    dt = __pd.DataFrame(__pd.to_timedelta(df[col_name] * 1000000000))
    decomp = dt[col_name].dt.components[['days', 'hours', 'minutes']]
    decomp['days'] = decomp['days'] % 7
    decomp.columns = ['{}_{}'.format(col_name, i) for i in decomp.columns]
    return __pd.concat([df, decomp], axis=1)


def get_pct_empty_df(df):
    return __pd.DataFrame({'Emptiness (%)': __np.round((df.isnull().sum() / df.shape[0]) * 100)})


def fill_cat_nan(df, col_name, fill_with='NA', return_df=True):
    tmp = df[col_name].apply(lambda x: fill_with if __pd.isnull(x) else x)
    if not return_df:
        return tmp
    df[col_name] = tmp
    # Exploiter une boucle for pour faie un fill_categorical sur un ensemble de colonnes
    return df


# Filling numerical with median
def fill_num_nan(df, cols):
    if type(cols) == str:
        cols = [cols]
    df_local = df.copy()
    for col in cols:
        fill_value = df[col].median()
        df_local[col].fillna(fill_value, inplace=True)  # fill with median because mean may be affect by outliers.
    return df_local


# Cleaning infinite values to NaN
def clean_inf_nan(df):
    """Useful after numpy calculation errors to remove infinite values
    """
    return df.replace([__np.inf, -__np.inf], __np.nan)  # replace all nan,inf,-inf to nan so it will be easy to replace


def get_dum(df, col_name, del_col=True, drop_first=True):
    dum = __pd.get_dummies(df[col_name], drop_first=drop_first)
    if del_col:
        df = df.drop([col_name], axis=1)
    dum.columns = ['{}_{}'.format(col_name, i) for i in dum.columns]
    df = __pd.concat([df, dum], axis=1)
    return df


def get_nb_post_point(x):
    if not '.' in str(x):
        return 0
    e = str(x).split('.')
    # if len(e[-1])>5:
    #    print(e)
    if e[-1] == '0':
        return 0
    return len(e[-1])


def get_nb_pre_point(x):
    e = str(x).split('.')
    if e[0] == '0':
        return 0
    return len(e[0])


def get_number_around_point(df, col_name):
    df['{}_pre'.format(col_name)] = df[col_name].apply(get_nb_pre_point)
    df['{}_post'.format(col_name)] = df[col_name].apply(get_nb_post_point)
    return df


def calculation_from_amount(df):
    df['TransactionAmt_squared'] = df['TransactionAmt'] ** 2
    df['TransactionAmt_log'] = __np.log(df['TransactionAmt'])
    return df


def label_encoding(df, cols, verbose=False, get_encoders=False):
    """
    """
    local_df = df.copy()

    # Checking if one columns or more
    if type(cols) == str:
        if verbose: print('There is only one column : {}'.format(cols))
        cols = [cols]

    # Encoding categorical values
    encoders = {}
    for col in cols:
        nb_unique = local_df[col].nunique()
        print("Label encoding {} with {} unique values".format(col, nb_unique))
        if local_df[col].dtype == 'object':
            local_df[col], encoders[col] = encode_serie(local_df[col])
        else:
            print('Error : {} is not categorical'.format(col))
    if get_encoders:
        return local_df, encoders
    return local_df


def spacify_number(number):
    """ Takes a number and returns a string with spaces every 3 numbers
    """
    nb_rev = str(number)[::-1]
    new_chain = ''
    for val, letter in enumerate(nb_rev):
        if val%3==0:
            new_chain += ' '
        new_chain += letter
    final_chain = new_chain[::-1]
    return final_chain


class log_wrapped_function(object):
    def __init__(self, function):
        self.function = function

    def log_and_call(self, *arguments, **namedArguments):
        __logging.info('>>> Function {}'.format(self.function.__name__))
        local = locals()
        if 'arguments' in local.keys():
            __logging.debug('- arguments      : {}'.format(local['arguments']))
        if 'namedArguments' in local.keys():
            __logging.debug('- namedArguments : {}'.format(local['namedArguments']))
        self.function.__call__(*arguments, **namedArguments)

def print_parameters(function):
    return log_wrapped_function(function).log_and_call


def list_path(path):
    # https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
    pass

if __name__ == "__main__":
    print("All functions have been loaded.")
    _ = get_functions_in_file('general.py', print_=True)



import numpy as __np
import pandas as __pd
import ruptures as __rpt
import matplotlib.pyplot as __plt

from sklearn.preprocessing import LabelEncoder as __LabelEncoder


def create_roll(ser_, nb_roll=100):
    ser_ = ser_.rolling(nb_roll, min_periods=1).mean()
    min_ser = min(ser_)
    max_ser = max(ser_ * 1.3)
    return ser_, min_ser, max_ser


def iqr_detect_outlier(data):
    """
    Detect outliers in data given Boxplot rule:
        X is an outlier <=> X>q3+1.5*IQR or X<q1-1.5*IQR 
    Args:
        data (Array): list of values to check outliers
        
    """

    q1, q3 = __np.percentile(data, [25, 75])
    iqr = q3 - q1
    cut_off = iqr * 1.5
    lower_bound, upper_bound = q1 - cut_off, q3 + cut_off

    return data[(data <= lower_bound) | (data >= upper_bound)]


def zscore_detect_outlier(data):
    """
    Detect outliers in data given Zscore rule:
        X is an outlier <=> |X|>E(X)+3*S(X)
    Args:
        data (Array): list of values to check outliers
        
    """
    threshold = 3
    mean = __np.mean(data)
    std = __np.std(data)
    z_score = __np.abs((data - mean) / std)
    outliers_idx = z_score[z_score > threshold].index

    return data.iloc[outliers_idx]


def shift_detect(data, model="l2", width=40, noise_std=4, debug=False):
    """
    Shift detection using window based method (see gitlab wiki for more info)
    
    Args:
        data (Array)    : list of values to check outliers
        model (String)  : which distance to use
        width (int)     : Window width
        noise_std(float): std for estimated noise
        debug (Bool)    : to display shift in data
        
     Returns:
         List: shift starting points 
    """

    n = len(data)
    pen = __np.log(n) * noise_std ** 2
    algo = __rpt.Window(width=width, model=model).fit(data)
    shifts = algo.predict(pen=pen)

    if debug:
        __rpt.show.display(data, shifts, figsize=(10, 6))
        __plt.show()

    return shifts[:-1]


def get_normalized_serie(df, col):
    normalized = ((df[col] - df[col].min()) / (df[col].max() - df[col].min()) + 1)
    return normalized


def encode_serie(serie):
    lbl = __LabelEncoder()
    lbl.fit(serie.apply(str).values)
    serie = lbl.transform(list(serie.apply(str).values))  # Ajout du str (sinon fonctionne pas sur nombre)
    return serie, lbl


# /Users/thibaud/Documents/Python_scripts/02_Projects/SNCF/open_data/sncf_utils.py
def transform_category_to_color(df, col_name, colors=None):
    if colors is None:
        colors = ['red', 'blue', 'green', 'yellow', 'orange'] * 50
    ser = df[col_name]
    ser, _ = encode_serie(ser)
    ser = __pd.Series(ser).apply(lambda x: colors[int(x)])
    return ser


def min_max_date(df):
    min_date = df['Year'].min()
    max_date = df['Year'].max()
    if min_date > max_date:
        tmp = min_date
        min_date = max_date
        max_date = tmp
    return min_date, max_date


def get_dummy(df, col):
    """
    Transform a pandas Series into dummies col
    """
    ser_ = df[col]
    dummies_data = __pd.get_dummies(ser_, drop_first=False)
    dummies_data.columns = ['{}_{}'.format(col, i) for i in dummies_data.columns]
    df = __pd.concat([df, dummies_data], axis=1)
    return df


def extract_car(df, col, car_number):
    """
    Extract one caracter of a pandas Series
    """
    new_col_name = '{}_{}'.format(col, str(car_number))
    ser_ = df[col].apply(str)
    df[new_col_name] = ser_.apply(lambda x: x[car_number])
    return df


def to_numeric_categorical(df, col):
    """
    Get a pandas Series and transforms every item into an integer value.
    """
    ser_ = df[col]
    unique_list = ser_.unique()
    nb = 0
    for i in unique_list:
        ser_ = ser_.apply(lambda x: x.replace(i, str(nb)))
        nb += 1
    df[col] = ser_
    return df


def get_categorical(df, verbose=False):
    cat_data = df.select_dtypes(include='object')
    num_data = df.select_dtypes(exclude='object')

    cat_cols = cat_data.columns.values
    num_cols = num_data.columns.values

    if verbose:
        print('\nCategorical Columns :\n ', cat_cols)
        print('\nNumerical Columns : \n', num_cols)
    return cat_cols, num_cols



def compute_variances(df, col, agg_col='ID_SEGMENT', round_=3):
    """
    About Variance https://bit.ly/3eM23Es
    - Variance inter : Variance of class's Average (variance résiduelle)
    - Variance intra : Weighted Average of class's Variances (variance expliquée)
    """

    # Variance
    var = np.var(df[col])
    var = round(var, round_)
    
    # Variance Inter
    var_inter = df.groupby(agg_col).agg({col:'mean'})
    var_inter = statistics.variance(var_inter[col])
    var_inter = round(var_inter, round_)
    
    # Variance Intra
    var_intra = df.groupby(agg_col).agg({col:[np.var, 'count']})
    var_intra.columns = ['var', 'count']
    var_intra['VAR_INTRA'] = (var_intra['count'] / var_intra['count'].sum()) * var_intra['var']
    var_intra = var_intra['VAR_INTRA'].round(round_).sum()

    return {
        'column':col,
        'var':var,
        'var_intra':var_intra,
        'var_inter':var_inter,
        'homogene':var_intra<var_inter
    }


from slacker import Slacker


def get_slack_token(path='slack.token'):
    with open(path, "r") as fichier:
        token = fichier.read()
    return token


def create_slack(slack_api_token):
    slack = Slacker(slack_api_token)
    return slack


def send_message(slack, message, channel='test_channel', username='Thib', emoji=':female-firefighter:'):
    slack.chat.post_message(
        channel=channel,
        text=message,
        username=username,
        icon_emoji=emoji
    )


if __name__ == "__main__":
    path = '/Users/thibaud/Documents/Python_scripts/Fonctions/slack.token'
    token = get_slack_token(path)
    slack = create_slack(token)
    send_message(slack, 'This is a test', channel='aléatoire')



import time
import numpy as np
import pandas as pd

from sklearn.model_selection import KFold
from sklearn.metrics import mean_absolute_error


def compute_log_score(X_test_type, y_test, y_pred):
    df = pd.concat([X_test_type.reset_index(drop=True), y_test.reset_index(drop=True), pd.Series(y_pred)], axis=1)
    df.columns = ['type', 'test', 'pred']
    score = 0
    for type_ in df.type.unique():
        tmp = df[df['type'] == type_]
        tmp['diff'] = tmp['test'] - tmp['pred']
        tmp['abs'] = np.abs(tmp['diff'])
        total = tmp['abs'].sum()
        moyenne = total / tmp.shape[0]
        moyenne_log = np.log(moyenne)
        score += moyenne_log
    score = score / df.type.nunique()
    return np.round(score, 4)


def cross_val_sk_learn(X, y, cv, model, score_metric=mean_absolute_error, round_=4):
    score_train = []
    score_test = []
    folds = KFold(n_splits=5, shuffle=True, random_state=20)

    for fold_n, (train_index, test_index) in enumerate(folds.split(X)):
        print('Fold', fold_n, 'started at', time.ctime())
        X_train, X_test = X.iloc[train_index], X.iloc[test_index]
        y_train, y_test = y.iloc[train_index], y.iloc[test_index]

        model.fit(X_train, y_train)

        y_pred_test = model.predict(X_test)
        y_pred_train = model.predict(X_train)

        train_score = np.round(score_metric(y_pred_train, y_train), round_)
        test_score = np.round(score_metric(y_pred_test, y_test), round_)
        # print('Score for train set: ', train_score)
        # print('Score for test set: ', test_score)
        score_train.append(train_score)
        score_test.append(test_score)
        # print('-----------------------------------------')
    print('Average score for train set: ', np.mean(score_train))
    print('Average score for test set: ', np.mean(score_test))
    print('Copyright Noctis')
    return score_train, score_test
